/*
 * Copyright © 2016 Vecly OÜ <info@vecly.com>
 *                  Karel Pärlin
 *                  mart.ruutel
 *                  Silver Kits
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; version 2 of the License.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along
 * with this program; if not, write to the Free Software Foundation, Inc.,
 * 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA.
 */

#include "m433-test-fw.h"
#include "rf.h"
#include "cc1111.h"

#include <string.h>

volatile __xdata uint8_t *rfTxBuffer;
volatile __xdata uint16_t rfTxCounter = 0;

volatile __xdata uint8_t rfRxBuffer[BUFFER_SIZE];
volatile __xdata uint16_t rfRxCounter = 0;
volatile __xdata uint16_t rfRxPackLen = 0;

volatile __xdata uint8_t usbWriteBuffer[BUFFER_SIZE];
volatile __xdata uint16_t usbWriteCounter = 0;

volatile __xdata uint8_t rfStatus = 0;

volatile __xdata uint8_t lastRssi = 0;

volatile __xdata uint8_t txPower = 7; // Initial power is max

void rfInit(void) {
    // init variables
    rfStatus = 0;
    memset(rfRxBuffer, 0, BUFFER_SIZE);
    rfRxCounter = 0;
    rfRxPackLen = 0;

    rfDefaultRegisters();

    // Some values are set here because we to not want to re set channel and Tx power.
    // Those can be set via control protocol
    // and there is no need to re set those to defaults every time.
    RF_CHANNR       = 0x00;
    RF_FREND0       = 0x17; // POWER

    // interrupts RFTXRX and RF priority 2
    IP0 = 0;
    IP1 = IP1_IPG0;

    RFTXRXIE = 0;
    RFIM = RFIM_IM_TXUNF | RFIM_IM_RXOVF | RFIM_IM_TIMEOUT | RFIM_IM_DONE | RFIM_IM_SFD;
    RFIF = 0;
    S1CON &= ~(S1CON_RFIF_0 | S1CON_RFIF_1);
    // IEN2 SFR is not divisible by 8 and therefore each
    // bit is not individually addressable as RFTXRXIE
    IEN2 |= IEN2_RFIE;

    RFST = RFST_SIDLE;
    while (MARCSTATE != RF_MARCSTATE_IDLE)
        ;
}

static void rfReset(void) {
    __xdata uint8_t s = MARCSTATE;
    RFST = RFST_SIDLE; while ((MARCSTATE) != RF_MARCSTATE_IDLE);

    switch(s) {
        case RF_MARCSTATE_RX:
            RFST = RFST_SRX;
            break;
        case RF_MARCSTATE_TX:
            RFST = RFST_STX;
            break;
        default:
            // In case reset to IDLE
            return;
    }
    
    while (MARCSTATE == RF_MARCSTATE_IDLE)
        ;
}

void rfTx(__xdata uint8_t *b, uint8_t l) {
    PKTLEN = l;

    rfTxBuffer = b;
    rfTxCounter = 0;

    // Stay in Idle after TX
    MCSM1 |= RF_MCSM1_TXOFF_MODE_IDLE;

    RFTXRXIE = 1;

    RFST = RFST_STX;

    while(MARCSTATE != RF_MARCSTATE_TX)
        ;
#ifdef __USBEVAL
    LED = 1;
#endif
#ifdef __M433LEDBOARD
    LED_RED = 1;
#endif
    while(MARCSTATE == RF_MARCSTATE_TX)
        ;
#ifdef __USBEVAL
    LED = 0;
#endif
#ifdef __M433LEDBOARD
    LED_RED = 0;
#endif

    // Clear pending interrupts.
    // Otherwise timer does not work until those are cleared.
    S1CON &= ~(S1CON_RFIF_0 | S1CON_RFIF_1);
}

void rfRx(void) {
    memset(rfRxBuffer, 0, BUFFER_SIZE);
    rfRxCounter = 0;
    rfRxPackLen = 0;
    RFTXRXIE = 1;

    S1CON &= ~(S1CON_RFIF_0 | S1CON_RFIF_1);
    RFIF &= ~RFIF_IM_DONE;

    // Stay in RX after RX
    MCSM1 |= RF_MCSM1_RXOFF_MODE_RX;

    RFST = RFST_SRX;
    while(MARCSTATE != RF_MARCSTATE_RX)
        ;
    return;
}

void rfIdle(void) {
    // Disable RFTXRX interrupt
    RFTXRXIE = 0;
    // Clear pending interrupts.
    // Otherwise timer does not work until those are cleared.
    S1CON &= ~(S1CON_RFIF_0 | S1CON_RFIF_1);

    MCSM1 = RF_MCSM1_CCA_MODE_RSSI_BELOW_UNLESS_RECEIVING | RF_MCSM1_RXOFF_MODE_IDLE | RF_MCSM1_TXOFF_MODE_IDLE;

    RFST = RFST_SIDLE;

    while(MARCSTATE != RF_MARCSTATE_IDLE)
        ;
    return;
}

void rfTxRxIsr(void) __interrupt RFTXRX_VECTOR {
    __xdata uint8_t i;

    // Clear interrupt flag
    RFTXRXIF = 0;
    S1CON &= ~(S1CON_RFIF_0 | S1CON_RFIF_1);

    if (MARCSTATE == RF_MARCSTATE_TX) {
        switch (operatingMode) {
            case 0x10:
                // We do not send data anyway when transmitting unmodulated carrier
                break;
            default:
                RFD = rfTxBuffer[rfTxCounter++];
                break;
        }

    } else if (MARCSTATE == RF_MARCSTATE_RX) {
#ifdef __USBEVAL
        LED = 1;
#endif
#ifdef __M433LEDBOARD
        LED_RED = 1;
#endif

        switch (operatingMode) {
            case 0x11:
                if (rfRxCounter == 1) {
                    lastRssi = RF_RSSI;
                    rfReset();
                    // Put some 0 to usb output buffer that we can send something
                    for( i = 0; i < 8; i++ )
                        usbWriteBuffer[ i ] = 0x30;
                    usbWriteCounter = 8;
                    rfRxCounter = 0;
                    return;
                }
                break;
            case 0x14:
                if (rfRxCounter == 1) {
                    // Message length is 1 byte bigger than packet length in first byte
                    rfRxPackLen = rfRxBuffer[0] + 1;
                }
                if (( rfRxPackLen > 0 ) && ( rfRxCounter == rfRxPackLen)) {
        #ifdef __USBEVAL
                    LED = 0;
        #endif
        #ifdef __M433LEDBOARD
                    LED_RED = 0;
        #endif
                    lastRssi = RF_RSSI;
                    rfReset();
                    // Move data to usb output buffer
                    for( i = 0; i < rfRxCounter; i++ )
                        usbWriteBuffer[ i ] = rfRxBuffer[ i ];
                    usbWriteCounter = rfRxPackLen;
                    rfRxCounter = 0;
                    rfRxPackLen = 0;
                    return;
                }
                break;
            default:
#ifdef __USBEVAL
                LED = 0;
#endif
#ifdef __M433LEDBOARD
                LED_RED = 0;
#endif
                break;
        }

        // Collect reveived bytes to Rx buffer
        rfRxBuffer[rfRxCounter++] = RFD;
    }
}

void rfIsr(void) __interrupt RF_VECTOR {
    // Clear interrupt flags
    S1CON &= (S1CON_RFIF_0 | S1CON_RFIF_1);

    if (RFIF & RFIF_IM_SFD) {
        // Sync word detected
        RFIF &= ~RFIF_IM_SFD;
    } if (RFIF & (RFIF_IM_DONE | RFIF_IM_TIMEOUT)) {
        // Packet received
        RFIF &= ~(RFIF_IM_DONE | RFIF_IM_TIMEOUT);
        rfStatus = RFIF_IM_DONE;
    } if (RFIF & RFIF_IM_RXOVF) {
        rfReset();
        rfStatus = RFIF_IM_RXOVF;
        RFIF &= ~RFIF_IM_RXOVF;
    } if (RFIF & RFIF_IM_TXUNF) {
        rfReset();
        rfStatus = RFIF_IM_TXUNF;
        RFIF &= ~RFIF_IM_TXUNF;
    }
}

void rfSetChannel(uint8_t ch) {
    // Must be done in IDLE
    rfIdle();
    switch (ch) {
    case 1:
        RF_CHANNR = 3;
        break;
    case 2:
        RF_CHANNR = 4;
        break;
    case 3:
        RF_CHANNR = 5;
        break;
    case 4:
        RF_CHANNR = 6;
        break;
    case 5:
        RF_CHANNR = 7;
        break;
    case 6:
        RF_CHANNR = 1;
        break;
    case 7:
        RF_CHANNR = 0;
        break;
    default:
        break;
     }
}

void rfSetTxPower(uint8_t p) {
    // values 0..7
    RF_FREND0 &= ~RF_FREND0_PA_POWER_MASK;
    RF_FREND0 |= RF_FREND0_PA_POWER_MASK & p;
    txPower = p;
}

void rfUnmodulatedContinuousTx(__xdata uint8_t tx) {
    // Registers and radio state is controlled via this function.
    // Goto idle first
    rfIdle();
    if (tx == 1){
        operatingMode = 0x10;
        safetyCountdownTimer = SAFETY_TIMER;
        // Needed radio registers except power, channel and frequency are replaced
        // with values generated in Smart RF studio.

        PKTCTRL1    = 0x04;
        PKTCTRL0    = 0x22;

        FSCTRL1     = 0x08;
        FSCTRL0     = 0x00;

        MDMCFG4     = 0xCA;
        MDMCFG3     = 0xA3;
        MDMCFG2     = 0x30;
        MDMCFG1     = 0x42;
        MDMCFG0     = 0x11;

        FOCCFG      = 0x16;
        BSCFG       = 0x6C;

        FREND0      = 0x11;

        FSCAL3      = 0xE9;
        FSCAL2      = 0x2A;
        FSCAL1      = 0x00;
        FSCAL0      = 0x1F;

        // TI requirement to set while on Tx.
        TEST2       = 0x88;
        TEST1       = 0x31;


        // OOK uses PA_TABLE1 and 0 for shifting.
        // Set both to same value to get unmodulated carrier.

        switch (txPower){
            case 0:
                RF_PA_TABLE1 = 0x0E;
                RF_PA_TABLE0 = 0x0E;
                break;
            case 1:
                RF_PA_TABLE1 = 0x1D;
                RF_PA_TABLE0 = 0x1D;
                break;
            case 2:
                RF_PA_TABLE1 = 0x34;
                RF_PA_TABLE0 = 0x34;
                break;
            case 3:
                RF_PA_TABLE1 = 0x2C;
                RF_PA_TABLE0 = 0x2C;
                break;
            case 4:
                RF_PA_TABLE1 = 0x60;
                RF_PA_TABLE0 = 0x60;
                break;
            case 5:
                RF_PA_TABLE1 = 0x84;
                RF_PA_TABLE0 = 0x84;
                break;
            case 6:
                RF_PA_TABLE1 = 0xC8;
                RF_PA_TABLE0 = 0xC8;
                break;
            case 7:
            default:
                RF_PA_TABLE1 = 0xC0;
                RF_PA_TABLE0 = 0xC0;
                break;
        }

        RFTXRXIE = 1;
        RFST = RFST_STX;
#ifdef __USBEVAL
        LED = 1;
#endif
#ifdef __M433LEDBOARD
        LED_RED = 1;
#endif

    } else if (tx == 0){
#ifdef __USBEVAL
        LED = 0;
#endif
#ifdef __M433LEDBOARD
        LED_RED = 0;
#endif
        // Clear pending interrupts.
        // Otherwise timer does not work until those are cleared.
        S1CON &= ~(S1CON_RFIF_0 | S1CON_RFIF_1);

        operatingMode = 0;
        safetyCountdownTimer = 0;
        // Set back initial values
        rfDefaultRegisters();
    }
}

void rfInitUnmodulatedRx(void){

    //Needed radio registers except power, channel and frequency are replaced
    //with values generated in Smart RF studio.
    PKTCTRL1    = 0x00;
    PKTCTRL0    = 0x00;

    FSCTRL1     = 0x08;
    FSCTRL0     = 0x00;

    // Channel spacing 99.975586 kHz
    // Data rate 38.3606 kBaud
    // Modulation format ASK/OOK
    // Rx filter 93.750000 kHz
    // NB! If you want receive only Rx with sync word switch MDMCFG2 to 0x32
    MDMCFG4     = 0xCA;
    MDMCFG3     = 0xA3;
    MDMCFG2     = 0x30;
    MDMCFG1     = 0x42;
    MDMCFG0     = 0x11;

    // Is it necessary to set state machine here?
    MCSM2       = 0x07;
    MCSM1       = 0x30;
    MCSM0       = 0x18;

    FOCCFG      = 0x16;
    BSCFG       = 0x6C;

    // Digital channel filter will detect signals roughly from - 80 dBm
    // This is set to avoid unnecessary background noise indication via
    // control protocol to client software.
    AGCCTRL2    = 0xC3;
    AGCCTRL1    = 0x40;
    AGCCTRL0    = 0x91;

    FREND1      = 0x56;

    FSCAL3      = 0xE9;
    FSCAL2      = 0x2A;
    FSCAL1      = 0x00;
    FSCAL0      = 0x1F;

    // TI recommendation for better sensitivity at low data rates.
    TEST2       = 0x81;
    TEST1       = 0x35;
}

void rfInitAskTx(void) {

    // Needed radio registers needed for Tx except power, channel and
    // frequency are replaced with values generated in Smart RF studio.

    // Append status to packet.
    // Add length as first byte and calculate CRC.
    // Needed to avoid CRC errors when receiving with TI Smart RF studio.
    PKTCTRL1    = 0x04;
    PKTCTRL0    = 0x05;

    FSCTRL1     = 0x08;
    FSCTRL0     = 0x00;

    // Channel spacing 99.975586 kHz
    // Data rate 38.3606 kBaud
    // Modulation format ASK/OOK
    // Manchester enabled
    // 30/32 sync word bits
    // 8 preamble bytes
    // Rx filter 93.750000 kHz
    MDMCFG4     = 0xCA;
    MDMCFG3     = 0xA3;
    MDMCFG2     = 0x3B;
    MDMCFG1     = 0x42;
    MDMCFG0     = 0x11;

    FOCCFG      = 0x16;
    BSCFG       = 0x6C;

    FREND0      = 0x11;

    FSCAL3      = 0xE9;
    FSCAL2      = 0x2A;
    FSCAL1      = 0x00;
    FSCAL0      = 0x1F;

    // TI requirement to set while on Tx.
    TEST2       = 0x88;
    TEST1       = 0x31;

    //OOK uses PA_TABLE1 and 0 for shifting.
    // PA_TABLE0 is 0 representation. Therefor it is set to 0.
    RF_PA_TABLE0 = 0x00;

    // Power is set from modem control protocol. 7 is default.
    switch (txPower){
        case 0:
            RF_PA_TABLE1 = 0x0E;
            break;
        case 1:
            RF_PA_TABLE1 = 0x1D;
            break;
        case 2:
            RF_PA_TABLE1 = 0x34;
            break;
        case 3:
            RF_PA_TABLE1 = 0x2C;
            break;
        case 4:
            RF_PA_TABLE1 = 0x60;
            break;
        case 5:
            RF_PA_TABLE1 = 0x84;
            break;
        case 6:
            RF_PA_TABLE1 = 0xC8;
            break;
        case 7:
        default:
            RF_PA_TABLE1 = 0xC0;
            break;
    }
}

void rfInitGfskTx(void) {

    // Needed radio registers needed for Tx except power, channel and
    // frequency are replaced with values generated in Smart RF studio.

    // Append status to packet.
    // Add length as first byte and calculate CRC.
    // Needed to avoid CRC errors when receiving with TI Smart RF studio.
    PKTCTRL1    = 0x04;
    PKTCTRL0    = 0x05;

    FSCTRL1     = 0x06;
    FSCTRL0     = 0x00;

    // Channel spacing 99.975586 kHz
    // Data rate 38.3606 kBaud
    // Modulation format GFSK
    // Manchester enabled
    // 30/32 sync word bits
    // 8 preamble bytes
    // Rx filter 93.750000 kHz
    MDMCFG4     = 0xCA;
    MDMCFG3     = 0xA3;
    MDMCFG2     = 0x1B;
    MDMCFG1     = 0x22;
    MDMCFG0     = 0x11;

    DEVIATN     = 0x36;

    FOCCFG      = 0x16;
    BSCFG       = 0x6C;

    FSCAL3      = 0xE9;
    FSCAL2      = 0x2A;
    FSCAL1      = 0x00;
    FSCAL0      = 0x1F;

    // TI requirement to set while on Tx.
    TEST2       = 0x88;
    TEST1       = 0x31;
}

void rfInitGfskRx(void) {

    // Needed radio registers needed for Tx except power, channel and
    // frequency are replaced with values generated in Smart RF studio.

    // Append status to packet.
    // Add length as first byte and calculate CRC.
    // Needed to avoid CRC errors when receiving with TI Smart RF studio.
    PKTCTRL1    = 0x04;
    PKTCTRL0    = 0x05;

    FSCTRL1     = 0x06;
    FSCTRL0     = 0x00;

    // Channel spacing 99.975586 kHz
    // Data rate 38.3606 kBaud
    // Modulation format GFSK
    // Manchester enabled
    // 30/32 sync word bits
    // 8 preamble bytes
    // Rx filter 93.750000 kHz
    MDMCFG4     = 0xCA;
    MDMCFG3     = 0xA3;
    MDMCFG2     = 0x1B;
    MDMCFG1     = 0x22;
    MDMCFG0     = 0x11;

    DEVIATN     = 0x36;

    FOCCFG      = 0x16;
    BSCFG       = 0x6C;

    RF_AGCCTRL2     = 0x43;
    RF_AGCCTRL1     = 0x40;
    RF_AGCCTRL0     = 0x91;

    FREND1      = 0x56;

    FSCAL3      = 0xE9;
    FSCAL2      = 0x2A;
    FSCAL1      = 0x00;
    FSCAL0      = 0x1F;

    // TI recommendation for better sensitivity at low data rates.
    TEST2       = 0x81;
    TEST1       = 0x35;
}

void rfDefaultRegisters(void) {
    RF_IOCFG2       = 0x00;
    RF_IOCFG1       = 0x00;
    RF_IOCFG0       = 0x00;
    RF_SYNC1        = 0xD3;
    RF_SYNC0        = 0x91;
    RF_PKTLEN       = 0xFF;
    RF_PKTCTRL1     = 0x04;
    RF_PKTCTRL0     = 0x45;
    RF_ADDR         = 0x00;

    RF_FSCTRL1      = 0x0c; // non default
    RF_FSCTRL0      = 0x00; // non default

    RF_FREQ2        = 0x12; // 433.5 MHz
    RF_FREQ1        = 0x10;
    RF_FREQ0        = 0x00;

    RF_MDMCFG4      = 0xca; // 100 kHz RX filter BW
    RF_MDMCFG3      = 0xb5; // non default
    RF_MDMCFG2      = 0x0e; // non default
    RF_MDMCFG1      = 0x42; // Channel spacing 100 kHz
    RF_MDMCFG0      = 0x11; // non default
    RF_DEVIATN      = 0x33; // Deviation 16 kHz
    RF_MCSM2        = 0x07;
    RF_MCSM1        = RF_MCSM1_CCA_MODE_RSSI_BELOW_UNLESS_RECEIVING | RF_MCSM1_RXOFF_MODE_IDLE | RF_MCSM1_TXOFF_MODE_IDLE;
    RF_MCSM0        = RF_MCSM0_FS_AUTOCAL_FROM_IDLE | RF_MCSM0_MAGIC_2;
    RF_FOCCFG       = 0x1d; // non default
    RF_BSCFG        = 0x1c; // non default
    RF_AGCCTRL2     = 0xc7; // non default
    RF_AGCCTRL1     = 0x00; // non default
    RF_AGCCTRL0     = 0xb0; // non default
    RF_FREND1       = 0xb6; // non default

    RF_FSCAL3       = 0xe9; // non default
    RF_FSCAL2       = 0x2a;
    RF_FSCAL1       = 0x00;
    RF_FSCAL0       = 0x1f;
    RF_TEST2        = 0x81; // non default
    RF_TEST1        = 0x35; // non default
    RF_TEST0        = 0x09; // non default
    RF_PA_TABLE7    = 0xc0; // 10 dBm
    RF_PA_TABLE6    = 0xc8; // 7 dBm
    RF_PA_TABLE5    = 0x84; // 5 dBm
    RF_PA_TABLE4    = 0x60; // 0 dBm
    RF_PA_TABLE3    = 0x2c; // -5 dBm
    RF_PA_TABLE2    = 0x34; // -10 dBm
    RF_PA_TABLE1    = 0x1d; // -15 dBm
    RF_PA_TABLE0    = 0x0e; // -20 dBm
}
